"""add workout description

Revision ID: 6988082918f8
Revises: c9c6a5a4dd6d
Create Date: 2024-09-07 20:02:56.341637

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6988082918f8'
down_revision = 'c9c6a5a4dd6d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.add_column(sa.Column('description', sa.String(length=10000), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workouts', schema=None) as batch_op:
        batch_op.drop_column('description')

    # ### end Alembic commands ###
